/***************************************************************************
    copyright            : (C) 2009 by Abdulhaq Lynch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; Version 2 of the License.               *
 *                                                                         *
 ***************************************************************************/

#include "calctime.h"
#include <QDate>
#include <QtDebug>

CalcTime::CalcTime()
{
}

void CalcTime::loadConfig(KPrayerTimeConfig & config)
{
    setMethod(config.m_itl_method);
    setPosition(config.m_latitude, config.m_longitude, config.m_local_elevation);
    setErrorCorrection(config.m_error_correction_min, config.m_error_clock_min);
    setTimeZone(config.m_time_zone);
    setZawaal(config.m_zawaal);
    setDST(config.m_dst);
}

void CalcTime::setMethod(int itlMethod)
{
  m_method_type = itlMethod;
}

void CalcTime::setPosition(double lat, double longt, int alt)
{
  m_location.degreeLat = lat;
  m_location.degreeLong = longt;
  m_location.seaLevel = alt;
  // default pressure/temp values
  m_location.pressure = 1010;
  m_location.temperature= 10;
  m_location.dst = 0;
}

void CalcTime::setDate(QDate date)
{
  m_date.day = date.day();
  m_date.month = date.month();
  m_date.year = date.year();
}

void CalcTime::setTimeZone(float tz)
{
  m_location.gmtDiff = tz;
}

void CalcTime::setDST(bool dst)
{
  m_dst = dst;
}

void CalcTime::setErrorCorrection(int error_corr_min, int error_clock_min)
{
  m_error_corr_min = error_corr_min;
  m_error_clock_min = error_clock_min;
}

void CalcTime::setZawaal(int zawaal)
{
  m_zawaal_min = zawaal;
}

void CalcTime::calc()
{

  getMethod(m_method_type, &m_method);

  // override a few method settings
  //m_method.extreme = 0; //no extreme latitude methods for now
  m_method.round = 3; //best rounding method (otherwise, with seconds < 30, maghrib shows some seconds before right time)

  // no offsets
  for (int i = 0; i < 6; i++) m_method.offList[i] = 0;
  m_method.offset = 0;

  /* Call the main function to fill the Prayer times array of structures */
  getPrayerTimes(&m_location, &m_method, &m_date, m_prayer_times);
  getNextDayFajr(&m_location, &m_method, &m_date, &m_next_fajr);

}

void CalcTime::accountForErrors()
{// do error correction for atmospheric variation (algorithm already takes this into account as best it can)

  //schedule[0] += m_error_corr_min / 60.0; not sure how atmospheric variation affects dawn light?
  m_prayer_times[1].minute -= m_error_corr_min;
  //schedule[2] += m_error_corr_min / 60.0; dhuhr time is (astronomically) exact so no need to correct
  m_prayer_times[3].minute += (m_error_corr_min / 2.0); // asr time is only affected approximately half as much
  m_prayer_times[4].minute += m_error_corr_min;
  m_prayer_times[5].minute += m_error_corr_min;
    // do zawaal waiting period

  m_prayer_times[2].minute += m_zawaal_min;

  for (int i = 0; i < 6; i++)
  {
    if (m_prayer_times[i].minute >= 60 && m_prayer_times[i].minute != 99 )
    {
      m_prayer_times[i].hour++;
      m_prayer_times[i].minute -= 60;
    }

    if (m_prayer_times[i].minute < 0)
    {
      m_prayer_times[i].hour--;
      m_prayer_times[i].minute += 60;
    }

    if (m_prayer_times[i].hour >= 24) m_prayer_times[i].hour -= 24;
    if (m_prayer_times[i].hour < 0) m_prayer_times[i].hour += 24;

  };
}

void CalcTime::accountForDST()
{
  if (m_dst) {
    m_prayer_times[0].hour += 1;
    m_prayer_times[1].hour += 1;
    m_prayer_times[2].hour += 1;
    m_prayer_times[3].hour += 1;
    m_prayer_times[4].hour += 1;
    m_prayer_times[5].hour += 1;
  }

}

void CalcTime::makeSchedule()
{
  calc();
  accountForDST();
  accountForErrors();
}

void CalcTime::makeScheduleTomorrow()
{
  m_prayer_times[0] = m_next_fajr;
}

