/***************************************************************************
    copyright            : (C) 2009 by Abdulhaq Lynch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; Version 2 of the License.               *
 *                                                                         *
 ***************************************************************************/

#include "painter.h"
#include <QtDebug>

#include <QBitmap>

QColor COLOUR_NOW(0xA0, 0xF0, 0x40);
QColor COLOUR_APPROACHING(0xFF, 0x99, 0x0);
QColor COLOUR_PLAIN(255, 159, 51);

float toDegrees(float radians) {
	return radians * 360 / (2 * M_PI);
}

QPainter * debugPainter;

void drawDot(QPointF & point, QColor colour)
{
	debugPainter->save();
	debugPainter->setBrush(colour);
	debugPainter->drawEllipse(point, 2, 2);
	debugPainter->restore();
}

void drawLine(QPointF & point, float gradient, QColor colour)
{
	float length = 200;
	debugPainter->save();
	debugPainter->setPen(colour);
	QPointF destination, destination1;
	destination.setX(point.x() + length);
	destination.setY(point.y() - (length * gradient));
	destination1.setX(point.x() - length);
	destination1.setY(point.y() + (length * gradient));
	debugPainter->drawLine(point, destination);
	debugPainter->drawLine(point, destination1);
	debugPainter->restore();
}

Painter::Painter(int width, int height)
{
	m_width = width;
	m_height = height;

	m_top_left_x = - (width / 1.2 / 2) + width / 10.0;
	m_top_left_y = width - (width / 1.2);
	m_diameter = width / 1.2;
	m_radius = m_diameter / 2.0;

	m_origin_x = m_top_left_x + m_radius;
	m_origin_y = m_top_left_y + m_radius;

	m_inner_pies = 0;

}

Painter::~Painter()
{
}

QString Painter::getStringTime(Prayer time, bool arabic_numerals)
{
  if (time.minute == 99 || time.hour >= 99) return QString("N/A");

  QString sHour = QString::number(time.hour);
  QString sMin = QString::number(time.minute);
  if (sMin.length() == 1)
    sMin = "0" + sMin;

  if (arabic_numerals)
  {
    sHour = convertToArabic(sHour);
    sMin = convertToArabic(sMin);
  }

  return sHour + ":" + sMin;
}

QString Painter::convertToArabic(QString numerals)
{
  const uint codeArabic0 = 0x0660;

  QString new_number;
  new_number = numerals.replace("0",QChar(codeArabic0));
  new_number.replace("1",QChar(codeArabic0 + 1));
  new_number.replace("2",QChar(codeArabic0 + 2));
  new_number.replace("3",QChar(codeArabic0 + 3));
  new_number.replace("4",QChar(codeArabic0 + 4));
  new_number.replace("5",QChar(codeArabic0 + 5));
  new_number.replace("6",QChar(codeArabic0 + 6));
  new_number.replace("7",QChar(codeArabic0 + 7));
  new_number.replace("8",QChar(codeArabic0 + 8));
  new_number.replace("9",QChar(codeArabic0 + 9));

  return new_number;
}

void Painter::getTangentialPoint(float radius, float angle, QPointF & point)
{
	float line_angle = angle + M_PI;
	point.setX(m_origin_x - radius * sin(line_angle));
	point.setY(m_origin_y + radius * cos(line_angle));
}

void Painter::paint(QPainter *p, Prayer schedule[], QTime current_time,
		KPrayerTimeConfig & config)
{

	float angles[6];

	SweepAngle sweep_angles[10];

    m_font_family = config.m_font_family;

	debugPainter = p;

	calcAngles(angles, schedule);

	calcSweepAngles(angles, sweep_angles);

	if (m_inner_pies == 0) drawInnerPies(angles);

	showInnerPies(p);

	drawTimes(p, schedule, angles, sweep_angles, config.m_arabic_numerals, current_time);

	drawTimeNumerals(p, current_time);

	drawSweep(p, current_time, sweep_angles);

	drawTimeArrow(p, current_time);

//	drawSettings(p, config);

	drawControls(p);

}

void Painter::drawControls(QPainter *p)
{
	p->setPen(Qt::white);
	p->setBrush(Qt::black);

	float spot_radius = m_radius * 0.1;
	m_animate_coords = QRectF(m_origin_x - spot_radius/2, m_origin_y - spot_radius/2,
			spot_radius, spot_radius);
	p->drawEllipse(m_animate_coords);
}

void Painter::drawSettings(QPainter *p, KPrayerTimeConfig & config)
{
    float x, y;
    QString text;
    QFont f(m_font_family);
    f.setPointSize(m_width/30);
    QFontMetrics fm(f);
    p->setPen(Qt::white);
    p->setBrush(Qt::blue);

    text = "Lat: " + QLocale().toString(config.m_latitude) + "      Long: " +
												QLocale().toString(config.m_longitude);
    QRectF text_rect = fm.tightBoundingRect(text);
    x = m_width - text_rect.width();
    y = m_height - text_rect.height();
    text_rect.translate(x - text_rect.height(), y);
    p->drawRoundedRect(text_rect, text_rect.height()/5.0, text_rect.height()/5.0);
    p->drawText(x, y, text);
}

SweepAngle::SweepAngle() {}

SweepAngle::SweepAngle(float start_angle, float end_angle,
		QColor start_colour, QColor half_colour, SweepDirection direction=FORWARD)
{
	m_start_angle = start_angle;
	m_end_angle = end_angle;
	m_start_colour = start_colour;
	m_half_colour = half_colour;
	m_direction = direction;
}

void Painter::calcSweepAngles(float * angles, SweepAngle sweep_angles[10])
{
	QColor go_full = Qt::green;
	QColor go_half = QColor(0, 120, 0);
	QColor prepare_full = QColor(0xff, 0x99, 0x33);
	QColor prepare_half = QColor(0xff, 0xcc, 0x99);

	if (angles[5] > M_PI || angles[5] < M_PI/2) angles[5] = M_PI;

	sweep_angles[0] = SweepAngle(0, angles[0],
			prepare_full, prepare_half, REVERSE);
	float angle_0_1 = angles[0] + 5 * (angles[1] - angles[0])/6;
	sweep_angles[1] = SweepAngle(angles[0], angle_0_1, go_full, go_half);
	sweep_angles[2] = SweepAngle(angles[1], angles[2],
			prepare_full, prepare_half, REVERSE);
	float angle_2_3 = angles[2] + 5 * (angles[3] - angles[2])/6;
	sweep_angles[3] = SweepAngle(angles[2], angle_2_3, go_full, go_half);
	sweep_angles[4] = SweepAngle(angle_2_3, angles[3],
			prepare_full, prepare_half, REVERSE);
	float angle_3_4 = angles[3] + 5 * (angles[4] - angles[3])/6;
	sweep_angles[5] = SweepAngle(angles[3], angle_3_4, go_full, go_half);
	sweep_angles[6] = SweepAngle(angle_3_4, angles[4],
			prepare_full, prepare_half, REVERSE);
	float angle_4_5 = angles[4] + (angles[5] - angles[4])/3;
	sweep_angles[7] = SweepAngle(angles[4], angle_4_5, go_full, go_half);
	sweep_angles[8] = SweepAngle(angle_4_5, angles[5],
			prepare_full, prepare_half, REVERSE);
	sweep_angles[9] = SweepAngle(angles[5], M_PI, go_full, go_half);
}

void Painter::calcAngles(float * angles, Prayer * schedule)
{
	float time_minutes;

	for (int c = 0; c < 6; c++) {
		Prayer prayer = schedule[c];
		time_minutes = (prayer.hour * 60) + prayer.minute;
		angles[c] = time_minutes / (24 * 60) * (M_PI);
//		 qDebug() << "angles[" << c << "] is " << angles[c];
	}
}

float angleForTime(QTime & time) {
	float time_minutes = (time.hour() * 60) + time.minute();
	return time_minutes / (24 * 60) * (M_PI);
}

void Painter::drawSweep(QPainter *p, QTime current_time, SweepAngle sweep_angles[10])
{
	p->save();

	float time_angle = angleForTime(current_time);

	for (int c=0; c < 10; c++) {
		SweepAngle sweep_angle = sweep_angles[c];
                if ((time_angle >= sweep_angle.m_start_angle) &&
				(time_angle < sweep_angle.m_end_angle)) {
			if (sweep_angle.m_direction == FORWARD)
				drawSweepShape(p, sweep_angle.m_start_angle, sweep_angle.m_end_angle,
					sweep_angle.m_start_colour, sweep_angle.m_half_colour);
			else
				drawSweepShape(p, sweep_angle.m_end_angle, sweep_angle.m_start_angle,
								sweep_angle.m_start_colour, sweep_angle.m_half_colour);
		}
	}

	p->restore();
}

void Painter::drawSweepShape(QPainter *p, float start_angle, float end_angle,
		QColor colour, QColor half_colour)
{
	QPointF a, b, c;

	float large_radius_factor = 1.15;
	float small_radius_factor = 1.08;

	getTangentialPoint(m_radius * small_radius_factor, start_angle, a);
	getTangentialPoint(m_radius * large_radius_factor, start_angle, b);
	getTangentialPoint(m_radius * large_radius_factor, end_angle, c);

	Arc arc(a, c, m_radius);

	QLinearGradient gradient(a, c);
	gradient.setColorAt(0, colour);
	gradient.setColorAt(0.3, half_colour);
	gradient.setColorAt(1, Qt::black);

	QBrush brush(gradient);
	p->setPen(Qt::NoPen);
	p->setBrush(brush);

	QPainterPath path;
	path.moveTo(a);
	path.lineTo(b);
	arcTo(path, start_angle, end_angle, m_radius *  large_radius_factor);
	if (end_angle > start_angle) {
		arcToWithCenter(path, arc.m_x, arc.m_y,
				arc.m_end_angle, arc.m_start_angle, arc.m_radius);
	}
	else {
		arcToWithCenter(path, arc.m_x, arc.m_y,
			arc.m_start_angle, arc.m_end_angle, arc.m_radius);
	}
	path.closeSubpath();
	p->drawPath(path);

}

void Painter::arcToWithCenter(QPainterPath & p, float x, float y,
		float from_angle, float to_angle, float radius)
{
	float top_left_x, top_left_y, diameter;
	diameter = radius * 2;
	top_left_x = x - radius;
	top_left_y = y - radius;
	p.arcTo(top_left_x, top_left_y, diameter, diameter,
			90 - toDegrees(from_angle), - toDegrees(to_angle - from_angle));
}

void Painter::arcTo(QPainterPath & p, float from_angle, float to_angle, float radius)
{
	arcToWithCenter(p, m_origin_x, m_origin_y, from_angle, to_angle, radius);
}

void Painter::drawTimes(QPainter *p, Prayer * schedule,
		float * angles, SweepAngle sweep_angles[10],
		bool arabic_numerals, QTime current_time)
{
	float angle;
	float text_align_left = m_top_left_x + m_radius * 2 * 1.2;
	bool offset_below_line;

	for (int i = 0; i < 6; i++) {

		angle = angles[i];

		DrawTimeMode draw_time_mode = getDrawTimeMode(i, sweep_angles, current_time);
		QString displayTime = getStringTime(schedule[i], arabic_numerals);
		if (i == 1 || i == 5) {
			offset_below_line = true;
		}
		else {
			offset_below_line = false;
		}
		drawSalatTime(p, angle, text_align_left, displayTime,
						draw_time_mode, offset_below_line);
	}
}

bool Painter::timeInSweep(QTime time, SweepAngle sweep_angle)
{
	float time_angle = angleForTime(time);
        if (time_angle >= sweep_angle.m_start_angle && time_angle < sweep_angle.m_end_angle)
		return true;
	else return false;
}

DrawTimeMode Painter::getDrawTimeMode(int ix, SweepAngle sweep_angles[10],
		QTime current_time)
{
	if (ix == 0 && timeInSweep(current_time, sweep_angles[0])) return APPROACHING;
	if (ix == 0 && timeInSweep(current_time, sweep_angles[1])) return NOW;
	if (ix == 2 && timeInSweep(current_time, sweep_angles[2])) return APPROACHING;
	if (ix == 2 && timeInSweep(current_time, sweep_angles[3])) return NOW;
	if (ix == 3 && timeInSweep(current_time, sweep_angles[4])) return APPROACHING;
	if (ix == 3 && timeInSweep(current_time, sweep_angles[5])) return NOW;
	if (ix == 4 && timeInSweep(current_time, sweep_angles[6])) return APPROACHING;
	if (ix == 4 && timeInSweep(current_time, sweep_angles[7])) return NOW;
	if (ix == 5 && timeInSweep(current_time, sweep_angles[8])) return APPROACHING;
	if (ix == 5 && timeInSweep(current_time, sweep_angles[9])) return NOW;
	return PLAIN;
}

void Painter::drawSalatTime(QPainter * p, float angle, float x,
		QString text, DrawTimeMode draw_time_mode, bool offset_below_line)
{

	float y;
	QColor text_colour(COLOUR_PLAIN);

	p->save();
        QFont time_font = QFont(m_font_family, m_width / 18, false);
	QFontMetrics fm = QFontMetrics(time_font);
	p->setFont(time_font);

	p->setPen(COLOUR_PLAIN);
	p->setBrush(Qt::black);

	QPointF a,b,c,d;

	getTangentialPoint(m_radius, angle, a);
	getTangentialPoint(m_radius * 1.08, angle, b);
	getTangentialPoint(m_radius * 1.15, angle, c);

	d = QPointF(c);
	d.setX(x);

	y = c.y();

	QRect text_rect = fm.tightBoundingRect(QString('X') + text + QString('X'));
	if (offset_below_line) {
			y += text_rect.height()/1.5;
			}
	if (draw_time_mode == APPROACHING || draw_time_mode == NOW) {
		if (draw_time_mode == APPROACHING) {
			p->setBrush(COLOUR_APPROACHING);
			p->setPen(COLOUR_PLAIN);
			text_colour = Qt::white;
		}
		else {
			p->setBrush(COLOUR_NOW);
			p->setPen(COLOUR_NOW);
			text_colour = Qt::black;
		}
		p->drawRoundedRect(x - fm.width('X'), y - text_rect.height() + 1,
				text_rect.width(), text_rect.height() * 1.2,
                text_rect.width()/10.0, text_rect.height()/5.0);
	}

	p->drawLine(b, c);
	p->drawLine(c, d);
	p->setPen(text_colour);
	float text_y = y + text_rect.height()/10.0;
    p->drawText(x, text_y, text);
	p->restore();
}

void Painter::showInnerPies(QPainter * p) {
	p->drawPixmap(0, 0, *m_inner_pies);
}

void Painter::drawInnerPies(float * angles)
{

    qDebug() << "draw inner pies";

	m_inner_pies = new QPixmap(m_width, m_height);
	// fill with alpha=0 (transparent)
	m_inner_pies->fill(QColor(0, 0, 0, 0));
	QPainter * p = new QPainter(m_inner_pies);

    p->setRenderHint(QPainter::SmoothPixmapTransform);
    p->setRenderHint(QPainter::Antialiasing);
	p->setRenderHint(QPainter::TextAntialiasing);

	QColor black = QColor(0, 0, 0);
	QColor grey = QColor(35, 25, 23);
	QColor yellow_1 = QColor(100, 100, 100);
	QColor yellow_2 = QColor(180, 180, 100);
	QColor yellow_3 = QColor(230, 230, 100);
	QColor yellow_4 = QColor(255, 255, 60);
	QColor orange_1 = QColor(255, 99, 51);

	QColor colors[8];

	colors[0] = black;
	colors[1] = grey;
	colors[2] = orange_1;
	colors[3] = yellow_4;
	colors[4] = yellow_4;
	colors[5] = orange_1;
	colors[6] = grey;
	colors[7] = black;

	float start_angle;
	float end_angle;
	float span_angle;
	float offset_start_angle;

	QPen pen = QPen();
	pen.setBrush(QBrush(Qt::NoBrush));
	p->setPen(pen);

	for (int c = 0; c < 7; c++) {
    	QConicalGradient conical_grad = QConicalGradient();
    	conical_grad.setCenter(m_origin_x, m_origin_y);
    	if (c==0) start_angle = 0;
    	else start_angle = toDegrees(angles[c-1]);
    	if (c==6) end_angle = 180;
    	else end_angle = toDegrees(angles[c]);
    	span_angle = end_angle - start_angle + 1;
    	offset_start_angle = start_angle;
    	conical_grad.setColorAt(0.0, colors[c+1]);
    	conical_grad.setColorAt(span_angle/360.0, colors[c]);
    	conical_grad.setAngle(90 - offset_start_angle - span_angle);
    	QBrush brush = QBrush(conical_grad);
    	p->setBrush(brush);
    	p->drawPie(m_top_left_x, m_top_left_y, m_diameter, m_diameter,
    			   (90 - offset_start_angle - span_angle) * 16, span_angle * 16);
	}

	delete p;

}

void Painter::drawTimeNumerals(QPainter *p, QTime current_time)
{

	p->save();

	p->setBrush(Qt::NoBrush);
	QColor line_color = QColor(250, 10, 10, 250);
	QPen pen = QPen(line_color);
	pen.setWidth(1);
	p->setPen(pen);

        QFont f(m_font_family);
	f.setStyleStrategy(QFont::ForceOutline);
	f.setPointSize(m_width/13);
	QFontMetrics fm(f);

	float hours[3] = {6, 12, 18};
	float angles[3] = {135, 90, 45};

	for (int c = 0; c < 3; c++) {

		p->save();

		QPainterPath text_path;
		QString text;
                text.setNum(hours[c]);
                QRect bounding_rect = fm.tightBoundingRect(text);
                float text_width = bounding_rect.width();
                float text_height = bounding_rect.height();
		text_path.addText(0, 0, f, text);
		p->translate(m_origin_x, m_origin_y);
		p->rotate(180 - angles[c]);
                p->translate( - (text_width / 2.0), - m_radius + text_height * 1.3);
		p->drawPath(text_path);

		p->restore();
	}

	p->restore();
}

void Painter::drawTimeArrow(QPainter *p, QTime current_time)
{
	QPointF a,b,c,d,e,f,g,h,i;

	float time_minutes = (current_time.hour() * 60) + current_time.minute();
	float angle = time_minutes / (24 * 60) * (M_PI);
	float offset_angle = 0.08;
	getTangentialPoint(m_radius * 1.05, angle, a);
	getTangentialPoint(m_radius * 0.95, angle + offset_angle, b);
	getTangentialPoint(m_radius * 0.95, angle - offset_angle, c);
	getTangentialPoint(m_radius * 0.95, angle + offset_angle * 0.6, d);
	getTangentialPoint(m_radius * 0.95, angle - offset_angle * 0.6, e);
	getTangentialPoint(m_radius * 0.7, angle - offset_angle * 0.6, f);
	getTangentialPoint(m_radius * 0.7, angle + offset_angle * 0.6, g);
	getTangentialPoint(m_radius * 0.6, angle, h);
	getTangentialPoint(m_radius * 0.4, angle, i);

	QColor line_color = QColor(250, 250, 250, 100);
	QColor fill_color = QColor(50, 50, 150, 180);
	QBrush brush = QBrush(fill_color);
	p->setBrush(brush);
	QPen pen = QPen(line_color);
	pen.setWidth(2);
	p->setPen(pen);

	p->drawLine(a, b);
	p->drawLine(a, c);
	p->drawLine(b, d);
	p->drawLine(c, e);
	p->drawLine(e, f);
	p->drawLine(d, g);
	p->drawLine(g, h);
	p->drawLine(f, h);
	p->drawLine(i, h);

	QPainterPath path;
	path.moveTo(a);
	path.lineTo(b);
	path.lineTo(d);
	path.lineTo(g);
	path.lineTo(h);
	path.lineTo(f);
	path.lineTo(e);
	path.lineTo(c);
	path.lineTo(a);
	p->drawPath(path);

}

Arc::Arc(QPointF a, QPointF b, float radius)
{
	/* Calculate the centre and start/end angles for an arc of given tangential points
	 * and radius.
	 */
	double alpha, delta;
	double gradient_a_b, gradient_a_b_normal;
	double od, ad, dp, op;
	QPointF d;
	d.setX((a.x() + b.x()) / 2.0);
	d.setY((a.y() + b.y()) / 2.0);

	gradient_a_b = - ((b.y() - a.y())) / (b.x() - a.x());
	gradient_a_b_normal =  - 1/gradient_a_b;

	delta = atan(gradient_a_b_normal);

	ad = sqrt(pow(d.x() - a.x(), 2) + pow(d.y() - a.y(), 2));
	od = sqrt(pow(radius, 2) - pow(ad, 2));
	alpha = asin(ad / radius);

	dp = sin(delta) * od;
	op = cos(delta) * od;

	m_x = d.x() - op;
	m_y = d.y() + dp;
	m_start_angle = M_PI/2 - (delta + alpha);
	m_end_angle = m_start_angle + 2 * alpha;
	m_radius = radius;
}
